#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646`- or :pep:`692`-compliant **unpack type hint** (i.e.,
``typing.Unpack[...]`` type hint subscripted by either a :pep:`646`-compliant
type variable tuple *or* :pep:`692`-compliant :class:`typing.TypedDict`
subclass) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: [PEP 646] Actually implement *SUPERFICIAL* type-checking support for PEP
#646-compliant tuple type hint unpacking:
#    Tuple[int, *Tuple[str, ...], str] - a tuple type where the first element is
#    guaranteed to be of type int, the last element is guaranteed to be of type
#    str, and the elements in the middle are zero or more elements of type str..
#
#Interestingly, even detecting accursed objects like "*Tuple[str, ...]" at
#runtime is highly non-trivial. They do *NOT* have a sane unambiguous type,
#significantly complicating detection. For some utterly inane reason, their type
#is simply the ambiguous "types.GenericAlias" type. That... wasn't what we were
#expecting *AT ALL*. For example, under Python 3.13:
#    # Note that Python *REQUIRES* unpacked tuple type hints to be embedded in
#    # some larger syntactic construct. So, just throw it into a list. This is
#    # insane, because we're only going to rip it right back out of that list.
#    # Blame the CPython interpreter. *shrug*
#    >>> yam = [*tuple[int, str]]
#
#    # Arbitrary unpacked tuple type hint.
#    >>> yim = yam[0]
#    >>> repr(yim)
#    *tuple[int, str]  # <-- gud
#    >>> type(yim)
#    <class 'types.GenericAlias'>  # <-- *TOTALLY NOT GUD. WTF, PYTHON!?*
#
#    # Now look at this special madness. The type of this object isn't even in
#    # its method-resolution order (MRO)!?!? I've actually never seen that
#    # before. The type of any object is *ALWAYS* the first item in its
#    # method-resolution order (MRO), isn't it? I... guess not. *facepalm*
#    >>> yim.__mro__
#    (<class 'tuple'>, <class 'object'>)
#
#    # So, "*tuple[int, str]" is literally both a tuple *AND* a "GenericAlias"
#    # at the same time. That makes no sense, but here we are. What are the
#    # contents of this unholy abomination?
#    >>> dir(yim)
#    ['__add__', '__args__', '__bases__', '__class__', '__class_getitem__',
#    '__contains__', '__copy__', '__deepcopy__', '__delattr__', '__dir__',
#    '__doc__', '__eq__', '__format__', '__ge__', '__getattribute__',
#    '__getitem__', '__getnewargs__', '__getstate__', '__gt__', '__hash__',
#    '__init__', '__init_subclass__', '__iter__', '__le__', '__len__', '__lt__',
#    '__mro_entries__', '__mul__', '__ne__', '__new__', '__origin__',
#    '__parameters__', '__reduce__', '__reduce_ex__', '__repr__', '__rmul__',
#    '__setattr__', '__sizeof__', '__str__', '__subclasshook__',
#    '__typing_unpacked_tuple_args__', '__unpacked__', 'count', 'index']
#
#    # Lotsa weird stuff there, honestly. Let's poke around a bit.
#    >>> yim.__args__
#    (<class 'int'>, <class 'str'>)  # <-- gud
#    >>> yim.__typing_unpacked_tuple_args__
#    (<class 'int'>, <class 'str'>)  # <-- gud, albeit weird
#    >>> yim.__unpacked__
#    True  # <-- *WTF!?!? what the heck is this nonsense?*
#    >>> yim.__origin__
#    tuple  # <-- so you lie about everything, huh?
#
#Basically, the above means that the only means of reliably detecting an
#unpacked tuple type hint at runtime is as follows:
#def is_pep646_hint_tuple_unpacked(obj: object) -> bool:
#    return (
#        obj.__class__ is types.GenericAlias and
#        #FIXME: Globalize this magic constant for efficiency. *shrug*
#        obj.__mro__ == (tuple, object) and
#        getattr(obj, '__unpacked__', None) is True
#        (
#            getattr(obj, '__args__', False) ==
#            getattr(obj, '__typing_unpacked_tuple_args__', True)
#        )
#    )
#
#That's super-inefficient *AND* fragile across Python versions, but... what you
#gonna do, huh? PEP 646 authors *REALLY* dropped the ball on this one, sadly.
#
#Of course, all of that only gets us to just detecting these accursed objects.
#We then need to actually *TYPE-CHECK* their contents, somehow. A few ideas:
#* A parent tuple hint can contain at most *ONE* unpacked child tuple hint. So,
#  we'll now need to record the number of unpacked child tuple hints that have
#  been previously handled and raise an exception if two or more are seen. Ugh!
#* If the child tuple hint being unpacked is variadic (i.e., it's last item is
#  "...") while the parent tuple hint containing that child tuple hint is
#  fixed-length, we're now in trouble. This is exactly the example shown above.
#  Supporting this requires possibly intense generalizations to our code
#  generation algorithm for tuple hints, which previously partitioned support
#  for fixed-length and variadic tuple hints into two separate logic paths. Now,
#  the fixed-length tuple hint code path will need to detect and handle unpacked
#  child variadic tuple hints. Kinda madness, honestly. We sigh. *sigh*

#FIXME: [PEP 692] *LOL*. "Tuple[*Ts] == Tuple[Unpack[Ts]] == Tuple[object]"
#after reduction, a fixed-length tuple hint. Clearly, however,
#"Tuple[Unpack[Ts]]" should instead be semantically equivalent to a variadic
#tuple hint: e.g.,
#    # This is what we want! Tuple[*Ts] == Tuple[Unpack[Ts]] ==
#    Tuple[Any, ...]
#
#See commentary in "data_pep646" for how to address this. *sigh*

#FIXME: [PEP 692] Actually implement deep type-checking support for PEP
#692-compliant unpack type hints of the form "**kwargs:
#typing.Unpack[UserTypedDict]". Doing so will *ALMOST CERTAINLY* necessitate a
#new logic pathway for dynamically generating type-checking code efficiently
#type-checking the passed variadic keyword argument dictionary "**kwargs"
#against that user-defined "UserTypedDict". Feasible, but non-trivial. *sigh*

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeDecorHintPep646Exception,
    BeartypeDecorHintPep692Exception,
)
from beartype.typing import Optional
from beartype._data.hint.datahintpep import Hint
from beartype._data.hint.pep.sign.datapepsigns import (
    HintSignTypedDict,
    HintSignTypeVarTuple,
)
from beartype._util.func.arg.utilfuncargiter import ArgKind
from beartype._util.hint.pep.utilpepget import get_hint_pep_args
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

# ....................{ REDUCERS                           }....................
#FIXME: Unit test us up, including:
#* Unsubscripted "typing.Unpack" type hints, which should be unconditionally
#  *PROHIBITED.* They signify nothing. "typing.Unpack" should *ALWAYS* be
#  subscripted by at least something.
def reduce_hint_pep646692_unpack(
    hint: Hint,
    arg_kind: Optional[ArgKind],
    exception_prefix: str,
    **kwargs
) -> Hint:
    '''
    Reduce the passed :pep:`646`- or :pep:`692`-compliant **unpack type hint**
    (i.e., ``typing.Unpack[...]`` type hint subscripted by either a
    :pep:`646`-compliant type variable tuple *or* :pep:`589`-compliant
    :class:`typing.TypedDict` subclass) to a more readily digestible type hint.

    Specifically, if the passed hint is:

    * A :pep:`646`-compliant unpack type hint subscripted by a
      :pep:`646`-compliant type variable tuple, this reducer effectively ignores
      this hint by reduction to the ignorable :class:`object` superclass (e.g.,
      from ``typing.Unpack[Ts]`` to simply ``object``). Although non-ideal,
      generating code type-checking these hints is sufficiently non-trivial to
      warrant a (hopefully) temporary delay in doing so properly. Note that:

      * Python itself unconditionally expands every unpacking of a
        :pep:`646`-compliant type variable tuple to this form of a
        :pep:`646`-compliant unpack type hint (e.g., from ``*Ts`` to
        ``typing.Unpack[Ts]``).
      * As a consequence of the prior note, :pep:`646`-compliant unpack type
        hints are as common as unpackings of :pep:`646`-compliant type variable
        tuples -- which is to say, increasingly common.

    * A :pep:`692`-compliant unpack type hint subscripted by a
      :pep:`589`-compliant :class:`typing.TypedDict` subclass, this reducer
      effectively ignores this hint by reduction to the ignorable
      :class:`object` superclass (e.g., from ``**kwargs:
      typing.Unpack[UserTypedDict]`` to simply ``**kwargs``). Although
      non-ideal, generating code type-checking these hints is sufficiently
      non-trivial to warrant a (hopefully) temporary delay in doing so properly.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : object
        :pep:`646`- or :pep:`692`-compliant unpack type hint to be reduced.
    arg_kind : Optional[ArgKind]
        Either:

        * If this hint annotates a parameter of some callable, that parameter's
          **kind** (i.e., :class:`.ArgKind` enumeration member conveying the
          syntactic class of that parameter, constraining how the callable
          declaring that parameter requires that parameter to be passed).
        * Else, :data:`None`.
    exception_prefix : str
        Human-readable substring prefixing exception messages raised by this
        reducer.

    All remaining keyword parameters are passed as is to the lower-level
    :func:`._reduce_hint_pep646_args_or_kwargs` reducer.

    Returns
    -------
    object
        Lower-level type hint currently supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintPep646Exception
        If this hint does *not* annotate a variadic keyword parameter and is
        *not* subscripted by a single :pep:`646`-compliant type variable tuple.
    BeartypeDecorHintPep692Exception
        If this hint annotates a variadic keyword parameter but is *not*
        subscripted by a single :pep:`589`-compliant :class:`typing.TypedDict`
        subclass.
    '''

    # ....................{ LOCALS                         }....................
    # Tuple of the single child hint subscripting this parent unpack hint. Note
    # that the "typing.Unpack" type hint factory has already pre-validated this
    # factory to accept at most one child type hint:
    #     >>> from typing import Unpack
    #     >>> Unpack['shaking', 'my', 'head']
    #     TypeError: typing.Unpack accepts only single type. Got ('shaking',
    #     'my', 'head').
    hints_child = get_hint_pep_args(hint)

    # Child hint subscripting this parent unpack hint.
    hint_child = hints_child[0]

    # Sign uniquely identifying this child hint if any *OR* "None".
    hint_child_sign = get_hint_pep_sign_or_none(hint_child)

    # ....................{ PEP 692                        }....................
    # Validate the passed variadic positional or keyword parameter type hint.

    # If this hint directly annotates a variadic keyword parameter, this is a
    # PEP 692- rather than 646-compliant hint. In this case...
    if arg_kind is ArgKind.VARIADIC_KEYWORD:
        # If this child hint is *NOT* a PEP 589-compliant "typing.TypeDict"
        # subclass, raise an exception.
        if hint_child_sign is not HintSignTypedDict:
            raise BeartypeDecorHintPep692Exception(
                f'{exception_prefix}PEP 692 unpack type hint {repr(hint)} '
                f'child type hint {repr(hint_child)} not '
                f'PEP 589 "typing.TypeDict" subclass.'
            )
        # Else, this child hint is a PEP 589-compliant "typing.TypeDict"
        # subclass.

        # Silently reduce to a noop by returning the ignorable "object"
        # superclass. While non-ideal, worky is preferable to non-worky.
        return object
    # Else, this hint does *NOT* directly annotate a variadic keyword parameter,
    # implying this is a PEP 646-compliant hint.

    # ....................{ PEP 646                        }....................
    # If this child hint is *NOT* a PEP 646-compliant "typing.TypeVarTuple"
    # object, raise an exception.
    if hint_child_sign is not HintSignTypeVarTuple:
        raise BeartypeDecorHintPep646Exception(
            f'{exception_prefix}PEP 646 unpack type hint {repr(hint)} '
            f'child type hint {repr(hint_child)} not '
            f'PEP 646 type variable tuple '
            f'(i.e., "typing.TypeVarTuple" object).'
        )
    # Else, this child hint is a PEP 646-compliant "typing.TypeVarTuple" object.

    # Silently reduce to a noop by returning the ignorable "object" superclass.
    # While non-ideal, worky is preferable to non-worky.
    return object



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`646`-compliant :obj:`typing.Self` **unit tests**.

This submodule unit tests :pep:`646` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.11.0')
def test_decor_pep646() -> None:
    '''
    Test :pep:`646` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.11 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype_test.a00_unit.data.pep.data_pep646 import (
        unit_test_decor_pep646)

    # ....................{ ASSERTS                        }....................
    # Defer to this external function, externalized to avoid "SyntaxError"
    # exceptions under older Python versions importing this test submodule.
    unit_test_decor_pep646()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-compliant type hint getters** (i.e., low-level callables
introspecting PEP-compliant type hints for associated metadata).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPepException
from beartype.roar._roarexc import _BeartypeUtilTypeException
from beartype.typing import (
    Any,
    Optional,
)
from beartype._data.hint.datahintpep import (
    Hint,
    HintOrNone,
)
from beartype._data.hint.datahinttyping import (
    TypeException,
)
from beartype._data.hint.pep.sign.datapepsignset import (
    HINT_SIGNS_ORIGIN_ISINSTANCEABLE,
)
from beartype._util.hint.pep.proposal.pep585 import (
    get_hint_pep585_generic_typevars,
    is_hint_pep585_generic_unsubbed,
)
from beartype._data.hint.datahinttyping import TupleTypeVars

# ....................{ GETTERS ~ args                     }....................
def get_hint_pep_args(hint: object) -> tuple:
    '''
    Tuple of the zero or more **child type hints** subscripting (indexing) the
    passed PEP-compliant type hint if this hint was subscripted *or* the empty
    tuple otherwise (i.e., if this hint is unsubscripted and is thus either an
    unsubscriptable type hint *or* a subscriptable type hint factory that is
    unsubscripted).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This getter should always be called in lieu of attempting to directly
    access the low-level** ``__args__`` **dunder attribute.** Various
    singleton objects defined by the :mod:`typing` module (e.g.,
    :attr:`typing.Any`, :attr:`typing.NoReturn`) fail to define this attribute,
    guaranteeing :class:`AttributeError` exceptions from all general-purpose
    logic attempting to directly access this attribute. Thus this function,
    which "fills in the gaps" by implementing this oversight.

    **This getter lies rarely due to subscription erasure** (i.e., the malicious
    destruction of child type hints by parent type hint factories at
    subscription time). Callers should *not* assume that the objects originally
    subscripting this hint are still accessible. Although *most* hints preserve
    their subscripted objects over their lifetimes, a small subset of edge-case
    hints erase those objects at subscription time. This includes:

    * :pep:`585`-compliant empty tuple type hints (i.e., ``tuple[()]``), which
      despite being explicitly subscripted erroneously erase that subscription
      at subscription time. This does *not* extend to :pep:`484`-compliant
      empty tuple type hints (i.e., ``typing.Tuple[()]``), which correctly
      preserve that subscripted empty tuple.

    **This getter lies less than the comparable**
    :func:`.get_hint_pep_typevars` **getter.** Whereas
    :func:`.get_hint_pep_typevars` synthetically propagates type variables from
    child to parent type hints (rather than preserving the literal type
    variables subscripting this type hint), this getter preserves the literal
    arguments subscripting this type hint if any. Notable cases where the two
    differ include:

    * Generic classes subclassing pseudo-superclasses subscripted by one or
      more type variables (e.g., ``class MuhGeneric(Generic[S, T])``).
    * Unions subscripted by one or more child type hints subscripted by one or
      more type variables (e.g., ``Union[str, Iterable[Tuple[S, T]]]``).

    Parameters
    ----------
    hint : object
        PEP-compliant type hint to be inspected.

    Returns
    -------
    tuple
        Either:

        * If this hint defines an ``__args__`` dunder attribute, the value of
          that attribute.
        * Else, the empty tuple.

    Examples
    --------
    .. code-block:: pycon

       >>> import typing
       >>> from beartype._util.hint.pep.utilpepget import (
       ...     get_hint_pep_args)
       >>> get_hint_pep_args(typing.Any)
       ()
       >>> get_hint_pep_args(typing.List[int, str, typing.Dict[str, str]])
       (int, str, typing.Dict[str, str])
    '''

    # Tuple of the zero or more child type hints subscripting this hint if
    # this hint defines of the "__args__" dunder attribute *OR* "None"
    # otherwise (i.e., if this hint fails to define this attribute).
    hint_args = getattr(hint, '__args__', None)

    # If this hint does *NOT* define this attribute, return the empty tuple.
    if hint_args is None:
        return ()
    # Else, this hint defines this attribute.
    #
    # If this hint is subscripted by zero child type hints, this hint only
    # superficially appears to be unsubscripted but was actually subscripted
    # by the empty tuple (e.g., "tuple[()]", "typing.Tuple[()]"). Why?
    # Because:
    # * Python 3.11 made the unfortunate decision of ambiguously conflating
    #   unsubscripted type hints (e.g., "tuple", "typing.Tuple") with type
    #   hints subscripted by the empty tuple, preventing downstream
    #   consumers from reliably distinguishing these two orthogonal cases.
    # * Python 3.9 made a similar decision but constrained to only PEP
    #   585-compliant empty tuple type hints (i.e., "tuple[()]"). PEP
    #   484-compliant empty tuple type hints (i.e., "typing.Tuple[()]")
    #   continued to correctly declare an "__args__" dunder attribute of
    #   "((),)" until Python 3.11.
    #
    # Disambiguate these two cases on behalf of callers by returning a
    # 1-tuple containing only the empty tuple (i.e., "((),)") rather than
    # returning the empty tuple (i.e., "()").
    elif not hint_args:
        return _HINT_ARGS_EMPTY_TUPLE
    # Else, this hint is either subscripted *OR* is unsubscripted but not
    # PEP 585-compliant.

    # In this case, return this tuple as is.
    return hint_args


# ....................{ GETTERS ~ typevars                 }....................
def get_hint_pep_typevars(hint: Hint) -> TupleTypeVars:
    '''
    Tuple of all **unique type variables** (i.e., subscripted :class:`TypeVar`
    instances of the passed PEP-compliant type hint listed by the caller at
    hint declaration time ignoring duplicates) if any *or* the empty tuple
    otherwise.

    This getter correctly handles both:

    * **Direct parametrizations** (i.e., cases in which this object itself is
      directly parametrized by type variables).
    * **Superclass parametrizations** (i.e., cases in which this object is
      indirectly parametrized by one or more superclasses of its class being
      directly parametrized by type variables).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This function should always be called in lieu of attempting to directly
    access the low-level** ``__parameters__`` **dunder attribute.** Various
    singleton objects defined by the :mod:`typing` module (e.g.,
    :attr:`typing.Any`, :attr:`typing.NoReturn`) fail to define this attribute,
    guaranteeing :class:`AttributeError` exceptions from all general-purpose
    logic attempting to directly access this attribute. Thus this function,
    which "fills in the gaps" by implementing this oversight.

    **Generics** (i.e., PEP-compliant type hints whose classes subclass one or
    more public :mod:`typing` pseudo-superclasses) are often but *not* always
    typevared. For example, consider the untypevared generic:

    .. code-block:: pycon

       >>> from typing import List
       >>> class UntypevaredGeneric(List[int]): pass
       >>> UntypevaredGeneric.__mro__
       (__main__.UntypevaredGeneric, list, typing.Generic, object)
       >>> UntypevaredGeneric.__parameters__
       ()

    Likewise, typevared hints are often but *not* always generic. For example,
    consider the typevared non-generic:

    .. code-block:: pycon

       >>> from typing import List, TypeVar
       >>> TypevaredNongeneric = List[TypeVar('T')]
       >>> type(TypevaredNongeneric).__mro__
       (typing._GenericAlias, typing._Final, object)
       >>> TypevaredNongeneric.__parameters__
       (~T,)

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    Tuple[TypeVar, ...]
        Either:

        * If this object defines a ``__parameters__`` dunder attribute, the
          value of that attribute.
        * Else, the empty tuple.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a PEP-compliant type hint
        parametrized by one or more type variables.

    Examples
    --------
    .. code-block:: pycon

       >>> import typing
       >>> from beartype._util.hint.pep.utilpepget import (
       ...     get_hint_pep_typevars)

       >>> S = typing.TypeVar('S')
       >>> T = typing.TypeVar('T')
       >>> class UserList(typing.List[T]): pass

       >>> get_hint_pep_typevars(typing.Any)
       ()
       >>> get_hint_pep_typevars(typing.List[int])
       ()
       >>> get_hint_pep_typevars(typing.List[T])
       (T)
       >>> get_hint_pep_typevars(UserList)
       (T)
       >>> get_hint_pep_typevars(typing.List[T, int, S, str, T])
       (T, S)
    '''

    # Value of the "__parameters__" dunder attribute on this object if this
    # object defines this attribute (e.g., is *NOT* a PEP 585-compliant
    # unsubscripted generic) *OR* "None" otherwise (e.g., is such a generic).
    hint_pep_typevars = getattr(hint, '__parameters__', None)

    # If this object defines *NO* such attribute, synthetically reconstruct
    # this attribute for PEP 585-compliant unsubscripted generics. Notably...
    if hint_pep_typevars is None:
        # Reconstruct this attribute as either...
        hint_pep_typevars = (
            # If this hint is a PEP 585-compliant unsubscripted generic, the
            # tuple of all type variables parametrizing all pseudo-superclasses
            # of this generic;
            get_hint_pep585_generic_typevars(hint)
            if is_hint_pep585_generic_unsubbed(hint) else
            # Else, this hint is *NOT* a PEP 585-compliant unsubscripted
            # generic. In this case, the empty tuple.
            ()
        )
    # Else, this object defines this attribute.

    # Return this attribute.
    return hint_pep_typevars

# ....................{ GETTERS ~ origin                   }....................
def get_hint_pep_origin(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    #FIXME: This should probably be a new "BeartypeDecorHintPepOriginException"
    #type, instead. But it's unclear whether users will even ever see this
    #exception. So, for now, laziness prevails. Huzzah! *sigh*
    exception_cls: TypeException = _BeartypeUtilTypeException,
    exception_prefix: str = '',
) -> Hint:
    '''
    **Unsafe origin object** (i.e., arbitrary object originating the passed
    PEP-compliant type hint but *not* necessarily an isinstanceable class such
    that all objects satisfying this hint are instances of this class)
    originating this hint if this hint originates from an origin *or*
    raise an exception otherwise (i.e., if this hint originates from *no*
    origin).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilTypeException`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    Hint
        Arbitrary object this hint originates from.

    Raises
    ------
    exception_cls
        If this hint originates from *no* object.
    '''

    # Origin originating this hint if any *OR* "None" otherwise.
    hint_origin = get_hint_pep_origin_or_none(hint)

    # If this hint does *NOT* originate from anything, raise an exception.
    if hint_origin is None:
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'originates from no other object.'
        )
    # Else, this hint originates from something.

    # Return that something.
    return hint_origin


def get_hint_pep_origin_or_none(hint: Hint) -> HintOrNone:
    '''
    **Unsafe origin object** (i.e., arbitrary object originating the passed
    PEP-compliant type hint but *not* necessarily an isinstanceable class such
    that all objects satisfying this hint are instances of this class)
    originating this hint if this hint originates from an origin *or*
    :data:`None` otherwise (i.e., if this hint originates from *no* origin).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **The high-level** :func:`get_hint_pep_origin_type_isinstanceable` getter
    should always be called in lieu of this low-level function.** Whereas the
    former is guaranteed to return either an isinstanceable class or
    :data:`None`, this getter enjoys no such guarantees and instead returns an
    arbitrary object that may or may not actually be an instanceable class.

    If this getter *must* be called, **this getter should always be called in
    lieu of attempting to directly access the low-level** ``__origin__``
    **dunder attribute.** Various :mod:`typing` objects either fail to define
    this attribute or define this attribute non-orthogonally, including objects:

    * Failing to define this attribute altogether (e.g., :attr:`typing.Any`,
      :attr:`typing.NoReturn`).
    * Defining this attribute to be their unsubscripted :mod:`typing` type hint
      factories (e.g., :attr:`typing.Optional`, :attr:`typing.Union`).
    * Defining this attribute to be their actual origin types.

    Since the :mod:`typing` module neither guarantees the existence of this
    attribute nor imposes a uniform semantic on this attribute when defined,
    that attribute is *not* safely directly accessible. Thus this getter, which
    "fills in the gaps" by implementing this oversight.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    Optional[Hint]
        Either:

        * If this hint originates from an arbitrary object, that object.
        * Else, :data:`None`.

    Examples
    --------
    .. code-block:: pycon

       >>> import typing
       >>> from beartype._util.hint.pep.utilpepget import (
       ...     get_hint_pep_origin_or_none)

       # This is sane.
       >>> get_hint_pep_origin_or_none(typing.List)
       list
       >>> get_hint_pep_origin_or_none(typing.List[int])
       list
       >>> get_hint_pep_origin_or_none(typing.Union)
       None
       >>> get_hint_pep_origin_or_none(typing.Union[int])
       None

       # This is insane.
       >>> get_hint_pep_origin_or_none(typing.Union[int, str])
       Union

       # This is crazy.
       >>> typing.Union.__origin__
       AttributeError: '_SpecialForm' object has no attribute '__origin__'

       # This is balls crazy.
       >>> typing.Union[int].__origin__
       AttributeError: type object 'int' has no attribute '__origin__'

       # This is balls cray-cray -- the ultimate evolution of crazy.
       >>> typing.Union[int, str].__origin__
       typing.Union
    '''

    # Return this hint's origin object if any *OR* "None" otherwise.
    return getattr(hint, '__origin__', None)

# ....................{ GETTERS ~ origin : type            }....................
#FIXME: Unit test us up, please.
def get_hint_pep_origin_type(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    #FIXME: This should probably be a new "BeartypeDecorHintPepOriginException"
    #type, instead. But it's unclear whether users will even ever see this
    #exception. So, for now, laziness prevails. Huzzah! *sigh*
    exception_cls: TypeException = _BeartypeUtilTypeException,
    exception_prefix: str = '',
) -> type:
    '''
    **Origin type** (i.e., class such that *all* objects satisfying the passed
    PEP-compliant type hint are instances of this class) originating this hint
    if this hint originates from such a type *or* raise an exception otherwise
    (i.e., if this hint originates from *no* such type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Type hint to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilTypeException`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    type
        Type originating this hint.

    Raises
    ------
    exception_cls
        If this hint either:

        * Does *not* originate from another object.
        * Originates from an object that is *not* a type.

    See Also
    --------
    :func:`.get_hint_pep_origin_or_none`
        Further details.
    '''

    # Origin type originating this hint if any *OR* "None" otherwise.
    hint_origin = get_hint_pep_origin_type_or_none(hint)

    # If *NO* origin type originates this hint...
    if hint_origin is None:
        assert isinstance(exception_cls, type), (
            f'{exception_cls} not exception type.')
        assert isinstance(exception_prefix, str), (
            f'{exception_prefix} not string.')

        # Origin non-type originating this hint if any *OR* "None" otherwise.
        hint_origin_nontype = get_hint_pep_origin_or_none(hint)

        # If this hint does *NOT* originate from another object, raise an
        # appropriate exception.
        if hint_origin_nontype is None:
            raise exception_cls(
                f'{exception_prefix}type hint {repr(hint)} '
                f'originates from no other object.'
            )
        # Else, this hint originates from another object. By definition, this
        # object *CANNOT* be a type.

        # Raise an appropriate exception.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} '
            f'originates from non-type {repr(hint_origin_nontype)}.'
        )
    # Else, this origin type originates this hint.

    # Return this origin type.
    return hint_origin


#FIXME: Unit test us up, please.
def get_hint_pep_origin_type_or_none(
    # Mandatory parameters.
    hint: Any,

    # Optional parameters.
    is_self_fallback: bool = False,
) -> Optional[type]:
    '''
    **Origin type** (i.e., class such that *all* objects satisfying the passed
    PEP-compliant type hint are instances of this class) originating this hint
    if this hint originates from such a type *or* :data:`None` otherwise (i.e.,
    if this hint originates from *no* such type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`.callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This high-level getter should always be called in lieu of either calling
    the low-level** :func:`.get_hint_pep_origin_or_none` **getter or attempting
    to directly access the low-level** ``__origin__`` **dunder attribute.**

    Parameters
    ----------
    hint : object
        Type hint to be inspected.
    is_self_fallback : bool = False
        :data:`True` only if returning the passed hint as a last-ditch fallback
        when this hint is a type defining the ``__origin__`` dunder attribute to
        be a non-type. Equivalently, if the passed hint is such a hint *and*
        this parameter is:

        * :data:`True`, this getter returns this hint as is.
        * :data:`False`, this getter returns :data:`None`.

        Defaults to :data:`False`. Explicit is better than implicit.

    Returns
    -------
    Optional[type]
        Either:

        * If this hint originates from a type, that type.
        * Else, :data:`None`.

    See Also
    --------
    :func:`.get_hint_pep_origin_or_none`
        Further details.
    '''
    assert isinstance(is_self_fallback, bool), (
        f'{repr(is_self_fallback)} not boolean.')

    # Unsafe origin type originating this hint if any *OR* "None" otherwise,
    # initialized to the arbitrary object set as the "hint.__origin__" dunder
    # attribute if this hint defines that attribute.
    hint_origin: Optional[type] = get_hint_pep_origin_or_none(hint)  # type: ignore[assignment]

    #FIXME: Unit test this up, please. *shrug*
    # If this origin is *NOT* a type...
    #
    # Ideally, this attribute would *ALWAYS* be a type for all possible
    # PEP-compliant type hints. For unknown reasons, type hint factories defined
    # by the standard "typing" module often set their origins to those same type
    # hint factories, despite those factories *NOT* being types. Why? Frankly,
    # we have no idea and neither does anyone else. Behold, true horror:
    #    >>> import typing
    #    >>> typing.Literal[1, 2].__origin__
    #    typing.Literal  # <-- do you even know what you are doing, python?
    #    >>> typing.Optional[int].__origin__
    #    typing.Union  # <-- wut? this is insane, python.
    if not isinstance(hint_origin, type):
        # Default this origin type to either...
        hint_origin = (
            # If...
            hint if (
                # The caller requests the "self" fallback logic *AND*...
                is_self_fallback and
                # This hint is itself a type, this hint could be euphemistically
                # said to originate from "itself." Fallback to this hint itself.
                # Look. Just go with it. We wave our hands in the air
                isinstance(hint, type)
            ) else
            # Else, either the caller did not request the "self" fallback logic
            # *OR* this hint is not a type. In either case, fallback to "None".
            None
        )
    # Else, this origin is a type.

    # Return this origin type.
    return hint_origin


def get_hint_pep_origin_type_isinstanceable(hint: Hint) -> type:
    '''
    **Isinstanceable origin type** (i.e., class passable as the second argument
    to the :func:`isinstance` builtin such that *all* objects satisfying the
    passed PEP-compliant type hint are instances of this class) originating
    this hint if this hint originates from such a type *or* raise an exception
    otherwise (i.e., if this hint does *not* originate from such a type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    type
        Standard origin type originating this hint.

    Raises
    ------
    BeartypeDecorHintPepException
        If this hint does *not* originate from a standard origin type.

    See Also
    --------
    :func:`get_hint_pep_origin_type_isinstanceable_or_none`
        Related getter.
    '''

    # Origin type originating this object if any *OR* "None" otherwise.
    hint_origin_type = get_hint_pep_origin_type_isinstanceable_or_none(hint)

    # If this type does *NOT* exist, raise an exception.
    if hint_origin_type is None:
        raise BeartypeDecorHintPepException(
            f'Type hint {repr(hint)} not isinstanceable (i.e., does not '
            f'originate from isinstanceable class).'
        )
    # Else, this type exists.

    # Return this type.
    return hint_origin_type


def get_hint_pep_origin_type_isinstanceable_or_none(
    hint: Hint) -> Optional[type]:
    '''
    **Standard origin type** (i.e., isinstanceable class declared by Python's
    standard library such that *all* objects satisfying the passed
    PEP-compliant type hint are instances of this class) originating this hint
    if this hint originates from such a type *or* :data:`None` otherwise (i.e.,
    if this hint does *not* originate from such a type).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This high-level getter should always be called in lieu of the low-level**
    :func:`get_hint_pep_origin_or_none` **getter or attempting to
    directly access the low-level** ``__origin__`` **dunder attribute.**

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    Optional[type]
        Either:

        * If this hint originates from a standard origin type, that type.
        * Else, :data:`None`.

    See Also
    --------
    :func:`get_hint_pep_origin_type_isinstanceable`
        Related getter.
    :func:`get_hint_pep_origin_or_none`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none

    # Sign uniquely identifying this hint if any *OR* "None" otherwise.
    hint_sign = get_hint_pep_sign_or_none(hint)

    # Return either...
    return (
        # If this sign originates from an origin type, that type;
        get_hint_pep_origin_or_none(hint)  # type: ignore[return-value]
        if hint_sign in HINT_SIGNS_ORIGIN_ISINSTANCEABLE else
        # Else, "None".
        None
    )

# ....................{ PRIVATE ~ args                     }....................
#FIXME: Shift into the "beartype._data.hint" subpackage somewhere, please.
_HINT_ARGS_EMPTY_TUPLE = ((),)
'''
Tuple containing only the empty tuple, to be returned from the
:func:`get_hint_pep_args` getter when passed either:

* A :pep:`585`-compliant type hint subscripted by the empty tuple (e.g.,
  ``tuple[()]``).
* A :pep:`484`-compliant type hint subscripted by the empty tuple (e.g.,
  ``typing.Tuple[()]``) under Python >= 3.11, which applied the :pep:`585`
  approach throughout the :mod:`typing` module.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Numba** integration tests.

This submodule functionally tests the :mod:`beartype` package against the
third-party Numba package.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('numba')
def test_numba_njit() -> None:
    '''
    Functional test validating that the :mod:`beartype` package raises *no*
    unexpected exceptions when decorating callables also decorated by the
    third-party :func:`numba.njit` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from numba import njit

    # ....................{ CALLABLES                      }....................
    @beartype
    @njit(cache=True)
    def had_shone_gleam_stony_orbs_so_from_his_steps() -> bool:
        '''
        Arbitrary callable decorated by both the :func:`beartype.beartype` *and*
        :func:`numba.njit` decorators in that order.
        '''

        return True

    # ....................{ ASSERTS                        }....................
    # Assert that calling the above callable returns the expected value.
    assert had_shone_gleam_stony_orbs_so_from_his_steps() is True



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) getters** (i.e., low-level callables
acquiring various properties of various nodes in the currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Module,
    dump as ast_dump,
    parse as ast_parse,
)
from beartype.roar._roarexc import _BeartypeUtilAstException

# ....................{ GETTERS ~ node                     }....................
#FIXME: Unit test us up, please.
def get_node_repr_indented(node: AST) -> str:
    '''
    Human-readable string pretty-printing the contents of the passed abstract
    syntax tree (AST), complete with readable indentation.

    Parameters
    ----------
    node : AST
        AST to be pretty-printed.

    Returns
    -------
    str
        Human-readable string pretty-printing the contents of this AST.
    '''
    assert isinstance(node, AST), f'{repr(node)} not AST.'

    # Return the pretty-printed contents of this AST.
    return ast_dump(node, indent=4)  # type: ignore[call-arg]

# ....................{ GETTERS ~ node                     }....................
#FIXME: Unit test us up, please. When we do, remove the "pragma: no cover" from
#the body of this getter below.
def get_code_child_node(code: str) -> AST:
    '''
    Abstract syntax tree (AST) node parsed from the passed (presumably)
    triple-quoted string defining a single child object.

    This function is principally intended to be called from our test suite as a
    convenient means of "parsing" triple-quoted strings into AST nodes.

    Caveats
    -------
    **This function assumes that this string defines only a single child
    object.** If this string defines either no *or* two or more child objects,
    an exception is raised.

    Parameters
    ----------
    code : str
        Triple-quoted string defining a single child object.

    Returns
    -------
    AST
        AST node encapsulating the object defined by this string.

    Raises
    -------
    _BeartypeUtilAstException
        If this string defines either no *or* two or more child objects.
    '''
    assert isinstance(code, str), f'{repr(code)} not string.'

    # "ast.Module" AST tree parsed from this string.
    node_module = ast_parse(code)

    # If this node is *NOT* actually a module node, raise an exception.
    if not isinstance(node_module, Module):  # pragma: no cover
        raise _BeartypeUtilAstException(
            f'{repr(node_module)} not AST module node.')
    # Else, this node is a module node.

    # List of all direct child nodes of this parent module name.
    nodes_child = node_module.body

    # If this module node contains either no *OR* two or more child nodes, raise
    # an exception.
    if len(nodes_child) != 1:  # pragma: no cover
        raise _BeartypeUtilAstException(
            f'Python code {repr(code)} defines '
            f'{len(nodes_child)} != 1 child objects.'
        )
    # Else, this module node contains exactly one child node.

    # Return this child node.
    return nodes_child[0]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`, :pep:`612`, and :pep:`644` **type parameter utility**
unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484612646` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep484612646_name() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484612646.get_hint_pep484612646_typeparam_name`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484612646Exception
    from beartype._data.hint.datahinttyping import T
    from beartype._util.hint.pep.proposal.pep484612646 import (
        get_hint_pep484612646_typeparam_name)
    from beartype._util.py.utilpyversion import (
        IS_PYTHON_AT_LEAST_3_11,
        IS_PYTHON_AT_LEAST_3_10,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter passed a PEP 484-compliant type variable returns the
    # name of this type variable.
    assert get_hint_pep484612646_typeparam_name(T) == 'T'

    # If the active Python interpreter targets Python >= 3.10 and thus supports
    # PEP 612...
    if IS_PYTHON_AT_LEAST_3_10:
        # Defer version-specific imports.
        from beartype.typing import ParamSpec

        # Arbitrary parameter specification.
        P = ParamSpec('P')

        # Assert this getter passed a PEP 612-compliant parameter specification
        # returns the name of this parameter specification.
        assert get_hint_pep484612646_typeparam_name(P) == 'P'

        # If the active Python interpreter targets Python >= 3.11 and thus
        # supports PEP 646...
        if IS_PYTHON_AT_LEAST_3_11:
            # Defer version-specific imports.
            from beartype.typing import TypeVarTuple

            # Arbitrary type variable tuple.
            Ts = TypeVarTuple('Ts')

            # Assert this getter passed a PEP 646-compliant type variable tuple
            # returns the name of this type variable tuple.
            assert get_hint_pep484612646_typeparam_name(Ts) == 'Ts'

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an object
    # that is *NOT* a type parameter.
    with raises(BeartypeDecorHintPep484612646Exception):
        get_hint_pep484612646_typeparam_name(
            'As with a palsied tongue, and while his beard')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable getters** (i.e., utility functions dynamically
querying and retrieving various properties of passed callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import (
    Callable,
    Optional,
)
from beartype._cave._cavefast import MethodBoundInstanceOrClassType
from beartype._data.hint.datahinttyping import (
    HintAnnotations,
    TypeException,
)

# ....................{ GETTERS ~ descriptors              }....................
#FIXME: Unit test us up, please.
#FIXME: Docstring us up, please.
def get_func_boundmethod_self(
    # Mandatory parameters.
    func: MethodBoundInstanceOrClassType,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> object:
    '''
    Instance object to which the passed **C-based bound instance method
    descriptor** (i.e., callable implicitly instantiated and assigned on the
    instantiation of an object whose class declares an instance function (whose
    first parameter is typically named ``self``) as an instance variable of that
    object such that that callable unconditionally passes that object as the
    value of that first parameter on all calls to that callable) was bound as an
    instance attribute at instantiation time.

    Parameters
    ----------
    func : object
        Bound method descriptor to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilCallableWrapperException`.
    exception_prefix : str, optional
        Human-readable label prefixing the representation of this object in the
        exception message. Defaults to the empty string.

    Returns
    -------
    object
        Instance object to which this bound method descriptor is bound to.

    Raises
    ------
    exception_cls
         If the passed object is *not* a bound method descriptor.

    See Also
    --------
    :func:`beartype._util.func.utilfunctest.is_func_boundmethod`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfunctest import die_unless_func_boundmethod

    # If this object is *NOT* a class method descriptor, raise an exception.
    die_unless_func_boundmethod(
        func=func,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this object is a class method descriptor.

    # Return the pure-Python function wrapped by this descriptor. Just do it!
    return func.__self__

# ....................{ GETTERS ~ hints                    }....................
#FIXME: Refactor all unsafe access of the low-level "__annotations__" dunder
#attribute to instead call this high-level getter, please.
#FIXME: See "FIXME:" comments in the "beartype._check.metadata.metadecor" submodule for
#how this needs to be refactored to support Python >= 3.13, please.
#FIXME: Unit test us up, please.
def get_func_annotations(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> HintAnnotations:
    '''
    **Annotations** (i.e., possibly empty ``__annotations__`` dunder dictionary
    mapping from the name of each annotated parameter or return of the passed
    possibly C-based callable to the type hint annotating that parameter or
    return) possibly annotating the passed callable if that callable defines the
    ``__annotations__`` dunder dictionary *or* raise an exception otherwise
    (i.e., if that callable fails to define the ``__annotations__`` dunder
    dictionary).

    Parameters
    ----------
    func : object
        Object to be inspected.
    exception_cls : TypeException, default: _BeartypeUtilCallableException
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilCallableException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintAnnotations
        Possibly empty ``__annotations__`` dunder dictionary defined by that
        callable.

    Raises
    ------
    exception_cls
         If that callable fails to define the ``__annotations__`` dunder
         dictionary. Since *all* pure-Python callables (including unannotated
         callables) define this dictionary, this getter raises an exception only
         if the passed callable is either uncallable *or* a **pseudo-callable
         object** (i.e., otherwise uncallable object whose class renders all
         instances of that class callable by defining the ``__call__()`` dunder
         method).

    See Also
    --------
    :func:`.get_func_annotations_or_none`
        Further details.
    '''

    # Annotations of that callable if that callable is actually a pure-Python
    # callable *OR* "None" otherwise.
    hint_annotations = get_func_annotations_or_none(func)

    # If that callable is *NOT* pure-Python, raise an exception.
    if hint_annotations is None:
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not class.')
        assert issubclass(exception_cls, Exception), (
            f'{repr(exception_cls)} not exception subclass.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # If that callable is uncallable, raise an appropriate exception.
        if not callable(func):
            raise exception_cls(f'{exception_prefix}{repr(func)} not callable.')
        # Else, that callable is callable.

        # Raise a human-readable exception.
        raise exception_cls(
            f'{exception_prefix}{repr(func)} not annotatable by type hints '
            f'(i.e., fails to define "__annotations__" dunder dictionary).'
        )
    # Else, that callable is pure-Python.

    # Return these annotations.
    return hint_annotations


#FIXME: Refactor all unsafe access of the low-level "__annotations__" dunder
#attribute to instead call this high-level getter, please.
#FIXME: Unit test us up, please.
def get_func_annotations_or_none(func: Callable) -> Optional[HintAnnotations]:
    '''
    **Annotations** (i.e., possibly empty ``__annotations__`` dunder dictionary
    mapping from the name of each annotated parameter or return of the passed
    possibly C-based callable to the type hint annotating that parameter or
    return) possibly annotating the passed callable if that callable defines the
    ``__annotations__`` dunder dictionary *or* :data:`None` otherwise (i.e., if
    that callable fails to define the ``__annotations__`` dunder dictionary).

    Parameters
    ----------
    func : object
        Object to be inspected.

    Returns
    -------
    Optional[HintAnnotations]
        Either:

        * If that callable is a pure-Python function, method, or object
          transitively wrapping a pure-Python function or method, the possibly
          empty ``__annotations__`` dunder dictionary defined by that callable.
        * Else, :data:`None`.
    '''

    # Demonstrable monstrosity demons!
    #
    # Note that:
    # * The "__annotations__" dunder attribute is guaranteed to exist *ONLY* for
    #   standard pure-Python callables. Various other callables of interest
    #   (e.g., functions exported by the standard "operator" module) do *NOT*
    #   necessarily declare that attribute. Since this getter is commonly called
    #   in general-purpose contexts where this guarantee does *NOT*
    #   necessarily hold, we intentionally access that attribute safely albeit
    #   somewhat more slowly via getattr().
    return getattr(func, '__annotations__', None)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **third-party module globals** (i.e., global constants broadly
concerning various third-party modules and packages rather than one specific
third-party module or package).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................

# ....................{ SETS                               }....................
#FIXME: Apply this blacklist to the following things:
#* Arbitrary callables to be decorated by @beartype, possibly. Consider defining
#  a new beartype._util.api.utilbeartype.is_func_thirdparty_blacklisted()
#  tester returning True *ONLY* if the passed callable has a "__module__" dunder
#  attribute whose value is a string residing in this frozenset.
THIRDPARTY_PACKAGE_NAMES_BLACKLIST = frozenset((
    # ....................{ ANTIPATTERN ~ forward ref      }....................
    # These third-party packages and modules widely employ the forward reference
    # antipattern throughout their codebases and are thus runtime-hostile.

    # Pydantic employs the forward reference antipattern everywhere: e.g.,
    #     from __future__ import annotations as _annotations
    #     import typing
    #     ...
    #
    #     if typing.TYPE_CHECKING:
    #         ...
    #         from ..main import BaseModel  # <-- undefined at runtime
    #
    #     ...
    #     def wrapped_model_post_init(
    #         self: BaseModel, context: Any, /) -> None:  # <-- unresolvable at runtime
    #
    # See also this @beartype-specific issue on Pydantic:
    #     https://github.com/beartype/beartype/issues/444
    'pydantic',

    # "urllib3" employs the forward reference antipattern everywhere: e.g.,
    #     import typing
    #     ...
    #
    #     if typing.TYPE_CHECKING:
    #         from typing import Final  # <-- undefined at runtime
    #
    #     ...
    #
    #     _DEFAULT_TIMEOUT: Final[_TYPE_DEFAULT] = _TYPE_DEFAULT.token
    #
    # See also this @beartype-specific comment on "urllib3":
    #     https://github.com/beartype/beartype/issues/223#issuecomment-2525261497
    'urllib3',

    # xarray employs the forward reference antipattern everywhere: e.g.,
    #     from __future__ import annotations
    #     from typing import IO, TYPE_CHECKING, Any, Generic, Literal, cast, overload
    #     ...
    #
    #     if TYPE_CHECKING:
    #         ...
    #         from xarray.core.dataarray import DataArray  # <-- undefined at runtime
    #
    #     ...
    #
    #     class Dataset(
    #         DataWithCoords,
    #         DatasetAggregations,
    #         DatasetArithmetic,
    #         Mapping[Hashable, "DataArray"],  # <-- unresolvable at runtime
    #     ):
    #
    # See also this @beartype-specific issue on xarray:
    #     https://github.com/beartype/beartype/issues/456
    'xarray',
))
'''
Frozen set of the fully-qualified names of all **beartype-blacklisted
third-party modules** well-known to be hostile to runtime type-checking in
general and :mod:`beartype` specifically, usually due to employing one or more
of the following antipatterns:

* The **forward reference antipattern,** a `harmful design pattern officially
  promoted throughout "mypy" documentation <antipattern_>`__. This antipattern
  leverages both :pep:`563` *and* the :pep:`484`-compliant
  :obj:`typing.TYPE_CHECKING` global (both of which are well-known to be hostile
  to runtime type-checking) to conditionally define relative forward references
  visible *only* to pure static type-checkers like ``mypy`` and ``pyright``.
  These references are undefined at runtime and thus inaccessible to hybrid
  runtime-static type-checkers like :mod:`beartype` and :mod:`typeguard`.

  As an example, consider this hypothetical usage of the forward reference
  antipattern in a mock third-party package named ``"awful_package"``:

  .. code-block:: python

     from __future__ import annotations  # <-- pep 563
     from typing import TYPE_CHECKING    # <-- pep 484

     if TYPE_CHECKING:                         # <---- "False" at runtime
         from awful_package import AwfulClass  # <-- undefined at runtime

     # PEP 563 (i.e., "from __future__ import annotations") stringifies the
     # undefined "AwfulClass" class to the string "'AwfulClass'" at runtime.
     # Since the "AwfulClass" class is undefined, however, neither @beartype nor
     # any other runtime type-checker can resolve this relative forward
     # reference to the external "awful_package.AwfulClass" class it refers to.
     def awful_func(awful_arg: AwfulClass): ...

.. _antipattern:
   https://mypy.readthedocs.io/en/latest/runtime_troubles.html#import-cycles
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) mungers** (i.e., low-level callables
modifying various properties of various nodes in the currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Call,
    ClassDef,
    Name,
    Subscript,
    expr,
    keyword,
)
from beartype.claw._clawmagic import (
    NODE_CONTEXT_LOAD,
    BEARTYPE_CLAW_STATE_OBJ_NAME,
    BEARTYPE_DECORATOR_FUNC_NAME,
)
from beartype.roar import BeartypeClawImportConfException
from beartype._conf.confmain import BeartypeConf
from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
from beartype._conf.confenum import BeartypeDecorationPosition
from beartype._data.hint.datahinttyping import NodeDecoratable
from beartype._util.ast.utilastmake import (
    make_node_kwarg,
    make_node_object_attr_load,
    make_node_str,
)
from beartype._util.ast.utilastmunge import copy_node_metadata

# ....................{ SUBCLASSES                         }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: To improve forward compatibility with the superclass API over which
# we have *NO* control, avoid accidental conflicts by suffixing *ALL* private
# and public attributes of this subclass by "_beartype".
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

class BeartypeNodeTransformerUtilityMixin(object):
    '''
    **Beartype abstract syntax tree (AST) node utility transformer** (i.e.,
    low-level mixin of the high-level
    :class:`beartype.claw._ast.clawastmain.BeartypeNodeTransformer` subclass
    supplementing that subclass with various low-level methods creating,
    modifying, and introspecting common node types and subclass properties).
    '''

    # ....................{ PRIVATE ~ decorators           }....................
    #FIXME: Unit test us up, please.
    def _decorate_node_beartype(
        self, node: NodeDecoratable, conf: BeartypeConf) -> None:
        '''
        Add a new **child beartype decoration node** (i.e., abstract syntax tree
        (AST) node applying the :func:`beartype.beartype` decorator configured
        by the passed beartype configuration) to the passed **parent decoratable
        node** (i.e., AST node encapsulating the definition of a pure-Python
        object supporting decoration by one or more ``"@"``-prefixed
        decorations, including both pure-Python classes *and* callables).

        Note that this function **prepends** rather than appends this child
        decoration node to the beginning of the list of all child decoration
        nodes for this parent decoratable node. Since that list is "stored
        outermost first (i.e. the first in the list will be applied last)",
        prepending guarantees that the beartype decorator will be applied last
        (i.e., *after* all other decorators). This ensures that explicitly
        configured beartype decorations applied to this decoratable by the end
        user (e.g., ``@beartype(conf=BeartypeConf(...))``) assume precedence
        over implicitly configured beartype decorations applied by this
        function.

        Parameters
        ----------
        node : AST
            **Decoratable node** (i.e., parent class or callable node) to add a
            new child beartype decoration node to.
        conf : BeartypeConf
            **Beartype configuration** (i.e., dataclass configuring the
            :mod:`beartype.beartype` decorator for some decoratable object(s)
            decorated by a parent node passing this dataclass to that
            decorator).
        '''
        assert isinstance(node, AST), f'{repr(node)} not AST node.'
        assert isinstance(conf, BeartypeConf), (
            f'{repr(conf)} not configuration.')

        # Child decoration node decorating that callable by the @beartype
        # decorator.
        beartype_decorator: expr = Name(
            id=BEARTYPE_DECORATOR_FUNC_NAME, ctx=NODE_CONTEXT_LOAD)

        # Copy all source code metadata from this parent callable node onto this
        # child decoration node.
        copy_node_metadata(node_src=node, node_trg=beartype_decorator)

        # If the current beartype configuration is *NOT* the default beartype
        # configuration, this configuration is a user-defined beartype
        # configuration which *MUST* be passed to a call to the @beartype
        # decorator. Merely referencing this decorator does *NOT* suffice. In
        # this case...
        if conf != BEARTYPE_CONF_DEFAULT:
            # Replace the reference to this decorator defined above with a
            # call to this decorator passed this configuration.
            beartype_decorator = Call(
                func=beartype_decorator,
                args=[],
                # Node encapsulating the passing of this configuration as
                # the "conf" keyword argument to this call.
                keywords=[
                    self._make_node_keyword_conf_beartype(node_sibling=node)],
            )

            # Copy all source code metadata from this parent callable node onto
            # this child call node.
            copy_node_metadata(node_src=node, node_trg=beartype_decorator)
        # Else, this configuration is simply the default beartype configuration.
        # In this case, avoid passing that configuration to the @beartype
        # decorator for both efficiency and simplicity.

        # Decorator position (i.e., location to which the @beartype decorator
        # will be implicitly injected into the existing chain of zero or more
        # decorators already decorating this parent decoratable node), defined
        # as either...
        decoration_position = (
            # If this parent decoratable node is a parent class node, the
            # class-specific decorator position.
            conf.claw_decoration_position_types
            if isinstance(node, ClassDef) else
            # Else, this parent decoratable node is *NOT* a parent class node.
            # By process of elimination, this parent decoratable node *MUST* be
            # a parent callable node. In this case, fallback to the
            # callable-specific decorator position.
            conf.claw_decoration_position_funcs
        )

        # If injecting the @beartype decorator last, prepend this child
        # decoration node to the beginning of the list of all child decoration
        # nodes for this parent decoratable node. Since this list is "stored
        # outermost first (i.e. the first in the list will be applied last)",
        # prepending guarantees that our decorator will be applied last (i.e.,
        # *AFTER* all subsequent decorators).
        if decoration_position is BeartypeDecorationPosition.LAST:
            node.decorator_list.insert(0, beartype_decorator)
        # If injecting the @beartype decorator first, append this child
        # decoration node to the end of the list of all child decoration nodes
        # for this parent decoratable node.
        elif decoration_position is BeartypeDecorationPosition.FIRST:
            node.decorator_list.append(beartype_decorator)
        # Else, an unrecognized decorator position was configured. In this case,
        # raise an exception. Note that this should *NEVER* occur.
        else:  # pragma: no cover
            raise BeartypeClawImportConfException(
                f'Beartype configuration {repr(conf)} '
                f'decorator position {repr(decoration_position)} unsupported.'
            )

    # ....................{ PRIVATE ~ factories            }....................
    #FIXME: Unit test us up, please.
    def _make_node_keyword_conf_beartype(self, node_sibling: AST) -> keyword:
        '''
        Create and return a new **beartype configuration keyword argument node**
        (i.e., abstract syntax tree (AST) node passing the beartype
        configuration associated with the currently visited module as a ``conf``
        keyword to a :func:`beartype.beartype` decorator orchestrated by the
        caller).

        Parameters
        ----------
        node_sibling : AST
            Sibling node to copy source code metadata from.

        Returns
        -------
        keyword
            Keyword node passing this configuration to an arbitrary function.
        '''

        # Node encapsulating the fully-qualified name of the current module.
        node_module_name = make_node_str(
            text=self._module_name_beartype, node_sibling=node_sibling)  # type: ignore[attr-defined]

        # Node encapsulating a reference to the beartype configuration object
        # cache (i.e., dictionary mapping from fully-qualified module names to
        # the beartype configurations associated with those modules).
        node_module_name_to_conf = make_node_object_attr_load(
            obj_name=BEARTYPE_CLAW_STATE_OBJ_NAME,
            attr_name='module_name_to_beartype_conf',
            node_sibling=node_sibling,
        )

        node_module_name_index: expr = None  # type: ignore[assignment]

        # Expression node encapsulating the indexation of a dictionary by the
        # fully-qualified name of the current module. For simplicity, simply
        # reuse this node.
        node_module_name_index = node_module_name

        # Node encapsulating a reference to this beartype configuration,
        # indirectly (and efficiently) accessed via a dictionary lookup into
        # this object cache. While cumbersome, this indirection is effectively
        # "glue" integrating this AST node generation algorithm with the
        # corresponding Python code subsequently interpreted by Python at
        # runtime during module importation.
        node_conf = Subscript(
            value=node_module_name_to_conf,
            slice=node_module_name_index,  # type: ignore[arg-type]
            ctx=NODE_CONTEXT_LOAD,
        )

        # Node encapsulating the passing of this beartype configuration as the
        # "conf" keyword argument to an arbitrary function call of some suitable
        # "beartype" function orchestrated by the caller.
        node_keyword_conf = make_node_kwarg(
            kwarg_name='conf', kwarg_value=node_conf, node_sibling=node_sibling)

        # Copy all source code metadata (e.g., line numbers) from this sibling
        # node onto these new nodes.
        copy_node_metadata(node_src=node_sibling, node_trg=node_conf)

        # Return this "conf" keyword node.
        return node_keyword_conf

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _is_scope_module_beartype(self) -> bool:
        '''
        :data:`True` only if the lexical scope of the currently visited node is
        the **module scope** (i.e., this node is declared directly in the body
        of the current user-defined module, implying this node to be a global).

        Returns
        -------
        bool
            :data:`True` only if the current lexical scope is a module scope.
        '''

        # Return true only if the stack of all lexical nodes is currently empty,
        # implying the current node resides directly in the body of a module.
        return not self._scope_stack_beartype  # type: ignore[attr-defined]


    @property
    def _is_scope_class_beartype(self) -> bool:
        '''
        :data:`True` only if the lexical scope of the currently visited node is
        a **class scope** (i.e., this node resides directly in the body of a
        user-defined class).

        Returns
        -------
        bool
            :data:`True` only if the current lexical scope is a class scope.
        '''

        # Return true only if...
        return (
            # The stack of all lexical scope is currently non-empty *AND*...
            bool(self._scope_stack_beartype) and  # type: ignore[attr-defined]
            # The current node resides directly in the body of a class.
            self._scope_stack_beartype[-1] is ClassDef  # type: ignore[attr-defined]
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **tuple type hint utilities**
(i.e., low-level callables generically applicable to both :pep:`484`- and
:pep:`585`-compliant tuple type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Tuple
from beartype._data.hint.pep.sign.datapepsigncls import HintSign
from beartype._data.hint.pep.sign.datapepsigns import (
    HintSignTuple,
    HintSignTupleFixed,
)
from beartype._util.hint.pep.proposal.pep484.pep484 import (
    HINT_PEP484_TUPLE_EMPTY)
from beartype._util.hint.pep.proposal.pep585 import (
    HINT_PEP585_TUPLE_EMPTY)

# ....................{ GETTERS                            }....................
#FIXME: Docstring us up, please.
#FIXME: Unit test us up, please.
def get_hint_pep484585_sign_tuplefixed_or_same(
    hint: object, hint_sign: HintSign) -> HintSign:
    '''
    The passed sign as is if this any sign other than the ambiguous
    :data:`.HintSignTuple` sign *or* the unambiguous :data:`.HintSignTupleFixed`
    sign if the passed hint is a fixed-length tuple hint.

    This low-level getter assists the higher-level
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_sign` getter to
    disambiguate the originally ambiguous :data:`.HintSignTuple` sign into:

    * If this hint is a fixed-length tuple hint, :data:`.HintSignTupleFixed`.
    * If this hint is a variable-length tuple hint, :data:`.HintSignTuple`.

    Parameters
    ----------
    hint : object
        PEP-compliant type hint to be inspected.
    hint_sign : HintSign
        Sign uniquely (but possibly ambiguously) identifying this hint.

    Returns
    -------
    HintSign
        Sign uniquely and unambiguously identifying this hint.
    '''
    assert isinstance(hint_sign, HintSign), f'{repr(hint_sign)} not sign.'

    # If this is a tuple hint, disambiguate between the following two
    # fundamentally distinct kinds of tuple hints:
    # * Fixed-length tuple type hints of the form
    #   "tuple[{hint_child_1}, ..., {hint_child_N}]", which this getter
    #   unambiguously reassigns the sign "HintSignTupleFixed".
    # * Variable-length tuple type hints of the form
    #   "tuple[{hint_child_1}, ...]", which this getter unambiguously
    #   preserves the sign "HintSignTuple".
    if hint_sign is HintSignTuple:
        # Avoid circular import dependencies.
        from beartype._util.hint.pep.utilpepget import get_hint_pep_args

        # Child hints subscripting this parent tuple hint.
        hint_childs = get_hint_pep_args(hint)
        # print(f'hint_childs: {hint_childs}')

        # Number of child hints subscripting this parent tuple hint.
        hint_childs_len = len(hint_childs)

        # Return the sign uniquely identifying either...
        return (
            # Variable-length tuple hints if either...
            HintSignTuple
            if (
                # This parent tuple hint is subscripted by *NO* child hints
                # and is thus the unsubscripted "typing.Tuple" type hint factory
                # semantically equivalent to the variable-length tuple hint
                # "typing.Tuple[object, ...]" *OR*...
                hint_childs_len == 0 or
                (
                    # This parent tuple hint is subscripted by exactly two child
                    # hints *AND*...
                    hint_childs_len == 2 and
                    # The second child hint is the ellipsis singleton (i.e.,
                    # the unquoted character sequence "...")...
                    hint_childs[1] is Ellipsis
                )
            ) else
            # Fixed-length tuple hints otherwise.
            HintSignTupleFixed
        )
    # Else, this is *NOT* a tuple hint.

    # Return this sign.
    return hint_sign

# ....................{ TESTERS                            }....................
def is_hint_pep484585_tuple_empty(hint: object) -> bool:
    '''
    :data:`True` only if the passed object is either a :pep:`484`- or
    :pep:`585`-compliant **empty fixed-length tuple type hint** (i.e., hint
    constraining objects to be the empty tuple).

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as this tester is only called under fairly
    uncommon edge cases.

    Motivation
    ----------
    Since type variables are not themselves types but rather placeholders
    dynamically replaced with types by type checkers according to various
    arcane heuristics, both type variables and types parametrized by type
    variables warrant special-purpose handling.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is an empty fixed-length tuple hint.
    '''

    # Return true only if this hint resembles either the PEP 484- or
    # 585-compliant fixed-length empty tuple type hint. Since there only exist
    # two such hints *AND* comparison against these hints is mostly fast, this
    # test is efficient in the general case.
    #
    # Note that this test may also be inefficiently performed by explicitly
    # obtaining this hint's sign and then subjecting this hint to specific
    # tests conditionally depending on which sign and thus PEP this hint
    # complies with: e.g.,
    #     # Return true only if this hint is either...
    #     return true (
    #         # A PEP 585-compliant "tuple"-based hint subscripted by no
    #         # child hints *OR*...
    #         (
    #             hint.__origin__ is tuple and
    #             hint_childs_len == 0
    #         ) or
    #         # A PEP 484-compliant "typing.Tuple"-based hint subscripted
    #         # by exactly one child hint *AND* this child hint is the
    #         # empty tuple,..
    #         (
    #             hint.__origin__ is Tuple and
    #             hint_childs_len == 1 and
    #             hint_childs[0] == ()
    #         )
    #     )
    return (
        hint == HINT_PEP585_TUPLE_EMPTY or
        hint == HINT_PEP484_TUPLE_EMPTY
    )

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_hint_pep484585_tuple_fixed_hint(hints: tuple) -> object:
    '''
    :pep:`484`- or :pep:`585`-compliant **fixed-length tuple type hint** of the
    form ``tuple[{hint_child1}, ???, {hint_childN}]`` subscripted by all
    PEP-compliant child type hints in the passed tuple.

    Parameters
    ----------
    hints : tuple
        Tuple of all child type hints to subscript this tuple type hint with.

    Returns
    -------
    object
        Fixed-length tuple type hint subscripted by these child type hints.
    '''
    assert isinstance(hints, tuple), f'{repr(hints)} not tuple.'

    # Return a fixed-length tuple type hint subscripted by these child type
    # hints, defined as either...
    return (
        #FIXME: Uncomment after dropping Python <= 3.10 support, which raises a
        #"SyntaxError" if we even try doing this. *SADNESS*
        # # If the active Python interpreter targets Python >= 3.11 and thus
        # # supports list unpacking in arbitrary expressions, prefer an efficient
        # # expression leveraging a list unpacking;
        # Tuple[*hints]
        # if IS_PYTHON_AT_LEAST_3_11 else
        # Else, the active Python interpreter targets Python <= 3.10.
        #
        # Dynamically subscript the builtin "tuple" type.
        Tuple.__class_getitem__(hints)  # type: ignore[attr-defined]
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype decorator :pep:`484`- and :pep:`585`-compliant unit tests.

This submodule unit tests :pep:`484` and :pep:`585` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeDecorHintPep585DeprecationWarning
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ decor : async              }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Urgh!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
async def test_decor_async_coroutine() -> None:
    '''
    Test decorating coroutines with the :func:`beartype.beartype` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from asyncio import sleep
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep585Exception
    from beartype_test._util.pytroar import raises_uncached
    from collections.abc import Coroutine as Pep585Coroutine
    from typing import Union, Coroutine as Pep484Coroutine

    # ....................{ CALLABLES                      }....................
    @beartype
    async def control_the_car(
        said_the: Union[str, int],
        biggest_greenest_bat: Union[str, float],
    ) -> Union[str, float]:
        '''
        Decorated coroutine whose return is annotated with an arbitrary
        PEP-compliant type hint.
        '''

        await sleep(0)
        return said_the + biggest_greenest_bat


    @beartype
    async def universal_love(
        said_the: Union[str, int], cactus_person: Union[str, float]) -> (
        Pep484Coroutine[None, None, Union[str, float]]):
        '''
        Decorated coroutine whose return is annotated with a
        :pep:`484`-compliant coroutine type hint.
        '''

        await sleep(0)
        return said_the + cactus_person


    @beartype
    async def transcendent_joy(
        said_the: Union[str, float], big_green_bat: Union[str, int]) -> (
        Pep585Coroutine[None, None, Union[str, float]]):
        '''
        Decorated coroutine whose return is annotated with a
        :pep:`585`-compliant coroutine type hint.
        '''

        await sleep(0)
        return said_the + big_green_bat

    # ....................{ PASS                           }....................
    # Assert awaiting this coroutine returns the expected value.
    assert await control_the_car(
        'I saw the big green bat bat a green big eye. ',
        'Suddenly I knew I had gone too far.') == (
        'I saw the big green bat bat a green big eye. '
        'Suddenly I knew I had gone too far.')

    # Assert awaiting this coroutine returns the expected value.
    assert await universal_love(
        'The sea was made of strontium; ', 'the beach was made of rye.') == (
        'The sea was made of strontium; the beach was made of rye.')

    # Assert awaiting this coroutine returns the expected value.
    assert await transcendent_joy(
        'A thousand stars of sertraline ',
        'whirled round quetiapine moons'
    ) == 'A thousand stars of sertraline whirled round quetiapine moons'

    # ....................{ FAIL                           }....................
    # Assert @beartype raises the expected exception when decorating a coroutine
    # whose return is annotated with a PEP 585-compliant coroutine type hint
    # *NOT* subscripted by exactly three child hints.
    with raises_uncached(BeartypeDecorHintPep585Exception):
        @beartype
        async def with_each_planck_moment_ever_fit() -> (
            Pep585Coroutine['to be eternally enjoyed']):
            await sleep(0)
            return 'Time will decay us but time can be left blank'


# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Hurk!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
async def test_decor_async_generator() -> None:
    '''
    Test decorating asynchronous generators with the :func:`beartype.beartype`
    decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from asyncio import sleep
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep484585Exception
    from beartype_test._util.pytroar import raises_uncached
    from beartype.typing import (
        AsyncGenerator as AsyncGeneratorUnsubscripted,
    )
    from collections.abc import (
        AsyncGenerator as Pep585AsyncGenerator,
        AsyncIterable as Pep585AsyncIterable,
        AsyncIterator as Pep585AsyncIterator,
    )
    from typing import (
        Union,
        AsyncGenerator as Pep484AsyncGenerator,
        AsyncIterable as Pep484AsyncIterable,
        AsyncIterator as Pep484AsyncIterator,
    )

    # ....................{ LOCALS                         }....................
    # Tuple of all PEP 484- or 585-compliant type hints annotating the returns
    # of asynchronous generators -- exercising all possible edge cases.
    HINTS_RETURN = (
        # ....................{ NON-PEP                    }....................
        AsyncGeneratorUnsubscripted,

        # ....................{ PEP 484                    }....................
        Pep484AsyncGenerator[Union[str, float], None],
        Pep484AsyncIterable[Union[str, float]],
        Pep484AsyncIterator[Union[str, float]],

        # ....................{ PEP 585                    }....................
        Pep585AsyncGenerator[Union[str, float], None],
        Pep585AsyncIterable[Union[str, float]],
        Pep585AsyncIterator[Union[str, float]],
    )

    # ....................{ PASS                           }....................
    # For each return type hint defined above...
    for hint_return in HINTS_RETURN:
        @beartype
        async def some_kind_of_spiritual_thing(
            said_the: Union[str, int], bigger_greener_bat: Union[str, float]
        ) -> hint_return:
            '''
            :func:`beartype.beartype`-decorated asynchronous generator whose
            return is annotated by a :pep:`484`- or :pep:`585`-compliant type
            hint deeply type-checking the value yielded by this generator.
            '''

            await sleep(0)
            yield said_the + bigger_greener_bat

        # Assert awaiting this asynchronous generator returns the expected
        # value. Unlike synchronous generators, asynchronous generators are
        # *NOT* actually iterators and thus have *NO* clean analogue to the
        # iter() and next() builtins. The closest approximation is this rather
        # unclean hack:
        #     await not_splitting_number.__anext__()
        #
        # See also this relevant StackOvelflow post:
        #     https://stackoverflow.com/a/42561322/2809027
        async for some_kind_of_spiritual in some_kind_of_spiritual_thing(
            'I should be trying to do some kind of spiritual thing ',
            'involving radical acceptance and enlightenment and such.',
        ):
            assert some_kind_of_spiritual == (
                'I should be trying to do some kind of spiritual thing '
                'involving radical acceptance and enlightenment and such.'
            )

    # ....................{ FAIL                           }....................
    # Assert this decorator raises the expected exception when decorating an
    # asynchronous generator annotating its return as anything *EXCEPT*
    # "AsyncGenerator[...]", "AsyncIterable[...]", or "AsyncIterator[...]".
    with raises_uncached(BeartypeDecorHintPep484585Exception):
        @beartype
        async def upside_down_trees(
            roots_in_the_breeze: str, branches_underground: str) -> str:
            await sleep(0)
            yield roots_in_the_breeze + branches_underground

# ....................{ TESTS ~ decor : sync               }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by the @beartype decorator below. Blargh!
@ignore_warnings(BeartypeDecorHintPep585DeprecationWarning)
def test_decor_sync_generator() -> None:
    '''
    Test decorating synchronous generators with the :func:`beartype.beartype`
    decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep484585Exception
    from beartype_test._util.pytroar import raises_uncached
    from beartype.typing import (
        Generator as GeneratorUnsubscripted,
    )
    from collections.abc import (
        Generator as Pep585Generator,
        Iterable as Pep585Iterable,
        Iterator as Pep585Iterator,
    )
    from typing import (
        Union,
        Generator as Pep484Generator,
        Iterable as Pep484Iterable,
        Iterator as Pep484Iterator,
    )

    # ....................{ LOCALS                         }....................
    # Tuple of all PEP 484- or 585-compliant type hints annotating the returns
    # of ynchronous generators -- exercising all possible edge cases.
    HINTS_RETURN = (
        # ....................{ NON-PEP                    }....................
        GeneratorUnsubscripted,

        # ....................{ PEP 484                    }....................
        Pep484Generator[Union[str, float], None, None],
        Pep484Iterable[Union[str, float]],
        Pep484Iterator[Union[str, float]],

        # ....................{ PEP 585                    }....................
        Pep585Generator[Union[str, float], None, None],
        Pep585Iterable[Union[str, float]],
        Pep585Iterator[Union[str, float]],
    )

    # ....................{ PASS                           }....................
    # For each return type hint defined above...
    for hint_return in HINTS_RETURN:
        @beartype
        def western_logocentric_stuff(
            said_the: Union[str, int], bigger_greener_bat: Union[str, float]
        ) -> hint_return:
            '''
            :func:`beartype.beartype`-decorated synchronous generator whose
            return is annotated by a :pep:`484`- or :pep:`585`-compliant type
            hint deeply type-checking the value yielded by this generator.
            '''

            yield said_the + bigger_greener_bat

        # Assert awaiting this synchronous generator yields the expected value
        # when iterated.
        assert next(western_logocentric_stuff(
              'all my Western logocentric stuff ', 'about factoring numbers',
        )) == 'all my Western logocentric stuff about factoring numbers'

    # ....................{ FAIL                           }....................
    # Assert this decorator raises the expected exception when decorating a
    # synchronous generator annotating its return as anything *EXCEPT*
    # "Generator[...]", "Iterable[...]", and "Iterator[...]".
    with raises_uncached(BeartypeDecorHintPep484585Exception):
        @beartype
        def GET_OUT_OF_THE_CAR(
            FOR_THE_LOVE_OF_GOD: str, FACTOR_THE_NUMBER: str) -> str:
            yield FOR_THE_LOVE_OF_GOD + FACTOR_THE_NUMBER
